from typing import Any, TypeVar

from multiprocessing import Manager

import torch
from torch.utils.data import Dataset

__all__ = ["CachedDataset"]


class NumpiedTensor:
    """
    A wrapper class that converts a PyTorch tensor to a NumPy array and vice versa.

    Args:
        tensor (torch.Tensor): The input PyTorch tensor.

    Attributes:
        array (np.ndarray): The NumPy array representation of the tensor.

    Methods:
        to_tensor: Converts the NumPy array back to a PyTorch tensor.
    """

    def __init__(self, tensor: torch.Tensor) -> None:
        self.array = tensor.numpy()

    def to_tensor(self) -> torch.Tensor:
        """
        Converts the NumPy array back to a PyTorch tensor.

        Returns:
            torch.Tensor: The PyTorch tensor representation of the NumPy array.
        """
        return torch.tensor(self.array)


def numpize_sample(sample: Any) -> Any:
    """
    Converts the input sample to a NumPy-compatible format.

    Args:
        sample (Any): The input sample to be converted.

    Returns:
        Any: The converted sample in a NumPy-compatible format.
    """

    if isinstance(sample, torch.Tensor):
        return NumpiedTensor(sample)
    elif isinstance(sample, tuple):
        return tuple(numpize_sample(s) for s in sample)
    elif isinstance(sample, list):
        return [numpize_sample(s) for s in sample]
    elif isinstance(sample, dict):
        return {k: numpize_sample(v) for k, v in sample.items()}
    else:
        return sample


def tensorize_sample(sample: Any) -> Any:
    """
    Converts the given sample into a tensor representation.

    Args:
        sample (Any): The input sample to be tensorized.

    Returns:
        Any: The tensorized representation of the input sample.
    """

    if isinstance(sample, NumpiedTensor):
        return sample.to_tensor()
    elif isinstance(sample, tuple):
        return tuple(tensorize_sample(s) for s in sample)
    elif isinstance(sample, list):
        return [tensorize_sample(s) for s in sample]
    elif isinstance(sample, dict):
        return {k: tensorize_sample(v) for k, v in sample.items()}
    else:
        return sample


T_co = TypeVar("T_co", covariant=True)


class CachedDataset(Dataset[T_co]):
    """
    A dataset wrapper that caches the samples to improve performance.

    Args:
        dataset (Dataset): The original dataset to be wrapped.

    Attributes:
        dataset (Dataset): The original dataset.
        manager (Manager): The multiprocessing manager.
        cache (dict): The cache dictionary to store the cached samples.
    """

    def __init__(self, dataset: Dataset[T_co]) -> None:
        self.dataset = dataset

        self.manager = Manager()
        self.cache = self.manager.dict()

    def __len__(self) -> int:
        return len(self.dataset)  # type: ignore[arg-type]

    def __getitem__(self, index: int) -> Any:
        if index not in self.cache:
            self.cache[index] = numpize_sample(self.dataset[index])

        return tensorize_sample(self.cache[index])



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from collections.abc import Callable, Iterator
from dataclasses import dataclass
from fnmatch import fnmatch

from torch import nn

import lightning as lt
from lightning.pytorch.utilities.types import OptimizerLRSchedulerConfig

from torch import optim


@dataclass
class OptimizerConfig:
    """
    Configuration for the optimizer and scheduler in a LightningModule.

    Args:
        optimizer_generator (Callable[..., optim.Optimizer]): A callable that generates the optimizer.
        params_transformer (Callable[..., Iterator[tuple[str, Any]]] | None): A callable that transforms the parameters
            into a format suitable for the optimizer. If None, the parameters are used as is. Defaults to None.
        scheduler_generator (Callable[..., optim.lr_scheduler.LRScheduler] | None):
            A callable that generates the learning rate scheduler. If None, no scheduler is used. Defaults to None.
        scheduler_interval (str | None): The interval at which the scheduler is called. Defaults to "step".
        scheduler_monitor (str | None): The metric to monitor for the scheduler. Defaults to "validation/loss".
    """

    optimizer_generator: Callable[..., optim.Optimizer]
    params_transformer: Callable[..., Iterator[tuple[str, Any]]] | None = None

    scheduler_generator: Callable[..., optim.lr_scheduler.LRScheduler] | None = None
    scheduler_interval: str | None = "step"
    scheduler_monitor: str | None = "validation/loss"


def build_param_groups(
    named_params: Iterator[tuple[str, nn.Parameter]],
    groups: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """
    Build parameter groups for the optimizer based on the provided patterns.

    Args:
        named_params (Iterator[tuple[str, nn.Parameter]]): An iterator of named parameters.
        groups (list[dict[str, Any]]): A list of dictionaries where each dictionary contains
            a "pattern" key that specifies the parameter names to match (``fnmatch``), and other optional keys.

    Example:
    In your config file, you might have:

    .. code-block:: yaml

        optimizer_config:
          _target_: aiaccel.torch.lightning.OptimizerConfig
          optimizer_generator:
            _partial_: True
            _target_: torch.optim.AdamW
            weight_decay: 0.01
          params_transformer:
              _partial_: True
              _target_: aiaccel.torch.lightning.build_param_groups
              groups:
                - pattern: "*bias"
                  lr: 0.01
                - pattern: "*weight"
                  lr: 0.001

    This will create two parameter groups: one for biases with a learning rate of 0.01 and another for weights with
    a learning rate of 0.001.
    """
    remaining = dict(named_params)

    param_groups = []
    for spec in groups:
        matched_params = []
        for target in [spec["pattern"]] if isinstance(spec["pattern"], str) else spec["pattern"]:
            matched_params += [remaining.pop(name) for name in list(remaining.keys()) if fnmatch(name, target)]

        assert len(matched_params) > 0

        param_groups.append({"params": matched_params} | {k: v for k, v in spec.items() if k != "pattern"})

    param_groups.append({"params": list(remaining.values())})

    return param_groups


class OptimizerLightningModule(lt.LightningModule):
    """
    LightningModule subclass for models that use custom optimizers and schedulers.

    Args:
        optimizer_config (OptimizerConfig): Configuration object for the optimizer.

    Attributes:
        optcfg (OptimizerConfig): Configuration object for the optimizer.

    Methods:
        configure_optimizers: Configures the optimizer and scheduler for training.
    """

    def __init__(self, optimizer_config: OptimizerConfig):
        super().__init__()

        self._optimizer_config = optimizer_config

    def configure_optimizers(self) -> optim.Optimizer | OptimizerLRSchedulerConfig:
        """
        Configures the optimizer and scheduler for training.

        Returns:
            Union[optim.Optimizer, OptimizerLRSchedulerConfig]: The optimizer and scheduler configuration.
        """

        params: Iterator[tuple[str, Any]] | Iterator[nn.Parameter]
        if self._optimizer_config.params_transformer is None:
            params = self.parameters()  # just because backward compatibility
        else:
            params = self._optimizer_config.params_transformer(self.named_parameters())

        optimizer = self._optimizer_config.optimizer_generator(params=params)  #

        if self._optimizer_config.scheduler_generator is None:
            return optimizer
        else:
            assert self._optimizer_config.scheduler_interval is not None
            assert self._optimizer_config.scheduler_monitor is not None
            return {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": self._optimizer_config.scheduler_generator(optimizer=optimizer),
                    "interval": self._optimizer_config.scheduler_interval,
                    "monitor": self._optimizer_config.scheduler_monitor,
                },
            }



#! /usr/bin/env python

from typing import Any

from argparse import ArgumentParser, Namespace
from pathlib import Path
import shlex
import subprocess
import sys
import time

import yaml

default_config_yaml = """\
walltime: "1:0:0"

script_prologue: |
    echo Job ID: $PBS_JOBID
    echo Hostname: $(hostname)

    export NVIDIA_VISIBLE_DEVICES=all

qsub: "qsub -P $JOB_GROUP -l walltime={args.walltime} -v USE_SSH=1"

cpu:
    qsub_args: "-q rt_HF -l select=1"
    job: "{command}"

cpu-array:
    n_tasks_per_proc: 128
    n_procs: 24
    qsub_args: "-q rt_HF -l select=1 -J 1-{args.n_tasks}:$(( {args.n_tasks_per_proc} * {args.n_procs} ))"
    job: "{command}"

gpu:
    qsub_args: "-q rt_HF -l select=1"
    job: "{command}"

gpu-array:
    n_tasks_per_proc: 128
    n_procs: 8
    qsub_args: "-q rt_HF -l select=1 -J 1-{args.n_tasks}:$(( {args.n_tasks_per_proc} * {args.n_procs} ))"
    job: "CUDA_VISIBLE_DEVICES=$(( LOCAL_PROC_INDEX % 8 )) {command}"

mpi:
    n_nodes: 1
    qsub_args: >-
        -q rt_HF
        -l select={args.n_nodes}:mpiprocs=$(( {args.n_procs} / {args.n_nodes} )):ompthreads=$(( {args.n_nodes} * 96 / {args.n_procs} ))
    job: |
        source /etc/profile.d/modules.sh
        module load hpcx

        mpirun -np {args.n_procs} -bind-to none -map-by slot \\
            -mca pml ob1 -mca btl self,tcp -mca btl_tcp_if_include bond0 \\
            {command}

train:
    qsub_args: >-
        -q $( (({args.n_gpus}==1)) && printf rt_HG || printf rt_HF )
        -l select=$(( ({args.n_gpus} + 7) / 8 )):mpiprocs=$( (({args.n_gpus}==1)) && printf 1 || printf 8 ):ompthreads=$( (({args.n_gpus}==1)) && printf 8 || printf 12 )
    job: |
        source /etc/profile.d/modules.sh
        module load hpcx

        mpirun -np {args.n_gpus} -bind-to none -map-by slot \\
            -mca pml ob1 -mca btl self,tcp -mca btl_tcp_if_include bond0 \\
            -x MAIN_ADDR=$(hostname -i) \\
            -x MAIN_PORT=3000 \\
            -x COLUMNS=120 \\
            -x PYTHONUNBUFFERED=true \\
            {command}
"""  # noqa: E501


def dispatch_job(mode: str, args: Namespace, config: Any) -> None:
    # Prepare the job script and arguments
    job = config[mode]["job"].format(command=shlex.join(args.command), args=args)

    if mode in ["cpu-array", "gpu-array"]:
        job = f"""\
for LOCAL_PROC_INDEX in {{1..{args.n_procs}}}; do
    TASK_INDEX=$(( PBS_ARRAY_INDEX + {args.n_tasks_per_proc} * (LOCAL_PROC_INDEX - 1) ))

    if [[ $TASK_INDEX -gt {args.n_tasks} ]]; then
        break
    fi

    TASK_INDEX=$TASK_INDEX \\
    TASK_STEPSIZE={args.n_tasks_per_proc} \\
        {job} > {args.log_filename.with_suffix("")}.${{PBS_ARRAY_INDEX}}-${{LOCAL_PROC_INDEX}}.log 2>&1 &

    pids[$LOCAL_PROC_INDEX]=$!
done

for i in "${{!pids[@]}}"; do
    wait ${{pids[$i]}}
done
"""
        job_log_filename = args.log_filename.with_suffix(".^array_index^.log")
        job_status_filename: Path = args.log_filename.with_suffix(".${PBS_ARRAY_INDEX}.out")

        status_filename_list = []
        for array_idx in range(0, args.n_tasks, args.n_tasks_per_proc * args.n_procs):
            status_filename_list.append(args.log_filename.with_suffix(f".{array_idx + 1}.out"))
    else:
        job_log_filename = args.log_filename
        job_status_filename = args.log_filename.with_suffix(".out")

        status_filename_list = [job_status_filename]

    job_script = f"""\
#! /bin/bash

#PBS -j oe
#PBS -k oed
#PBS -o {job_log_filename}

set -eE -o pipefail
trap 'echo $? > {job_status_filename}' ERR EXIT  # at error and exit
trap 'echo 143 > {job_status_filename}' TERM  # at termination (by job scheduler)

if [ -n "$PBS_O_WORKDIR" ] && [ "$PBS_ENVIRONMENT" != "PBS_INTERACTIVE" ]; then
    cd $PBS_O_WORKDIR
fi


{config["script_prologue"]}

{job}
"""

    qsub = config["qsub"].format(args=args)
    qsub_args = config[mode]["qsub_args"].format(args=args)

    # Create the job script file, remove old status files, and run the job
    args.log_filename.parent.mkdir(exist_ok=True, parents=True)

    job_filename: Path = args.log_filename.with_suffix(".sh")
    with open(job_filename, "w") as f:
        f.write(job_script)

    for status_filename in status_filename_list:
        status_filename.unlink(missing_ok=True)

    subprocess.run(f"{qsub} {qsub_args} {job_filename}", shell=True, check=True)

    for status_filename in status_filename_list:
        while not status_filename.exists():
            time.sleep(1.0)

        status = int(status_filename.read_text())
        if status != 0:
            raise RuntimeError(f"Job failed with {status} exit code.")
        status_filename.unlink()


def main() -> None:
    # Load configuration (from the default YAML string)
    parser = ArgumentParser(add_help=False)
    parser.add_argument("--print_config", action="store_true")
    parser.add_argument("--config", type=Path)
    args, _ = parser.parse_known_args()

    if args.config is not None:
        with open(args.config) as f:
            config_yaml = f.read()
    else:
        config_yaml = default_config_yaml

    if args.print_config:
        print(config_yaml)
        sys.exit(0)

    config = yaml.safe_load(config_yaml)

    # Parse command-line arguments
    parser = ArgumentParser()
    parser.add_argument("--print_config", action="store_true")
    parser.add_argument("--config", type=Path)
    sub_parsers = parser.add_subparsers(dest="mode", required=True)

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument("--walltime", type=str, default=config["walltime"])
    parent_parser.add_argument("log_filename", type=Path)
    parent_parser.add_argument("command", nargs="+")

    sub_parser = sub_parsers.add_parser("cpu", parents=[parent_parser])
    sub_parser.add_argument("--n_tasks", type=int)
    sub_parser.add_argument("--n_tasks_per_proc", type=int, default=config["cpu-array"]["n_tasks_per_proc"])
    sub_parser.add_argument("--n_procs", type=int, default=config["cpu-array"]["n_procs"])

    sub_parser = sub_parsers.add_parser("gpu", parents=[parent_parser])
    sub_parser.add_argument("--n_tasks", type=int)
    sub_parser.add_argument("--n_tasks_per_proc", type=int, default=config["gpu-array"]["n_tasks_per_proc"])
    sub_parser.add_argument("--n_procs", type=int, default=config["gpu-array"]["n_procs"])

    sub_parser = sub_parsers.add_parser("mpi", parents=[parent_parser])
    sub_parser.add_argument("--n_procs", type=int, required=True)
    sub_parser.add_argument("--n_nodes", type=int, default=config["mpi"]["n_nodes"])

    sub_parser = sub_parsers.add_parser("train", parents=[parent_parser])
    sub_parser.add_argument("--n_gpus", type=int)

    args = parser.parse_args()
    mode = args.mode + "-array" if getattr(args, "n_tasks", None) is not None else args.mode

    dispatch_job(mode, args, config)


if __name__ == "__main__":
    main()



from typing import Any, TypeVar

from multiprocessing import Manager
from pathlib import Path
import pickle as pkl
import uuid

from torch.utils.data import Dataset

__all__ = ["FileCachedDataset"]


T_co = TypeVar("T_co", covariant=True)


class FileCachedDataset(Dataset[T_co]):
    """
    A dataset wrapper that caches samples to disk to reduce memory usage.

    This class wraps an existing `torch.utils.data.Dataset` and caches samples
    as pickle files in a specified directory.

    Args:
        dataset (Dataset[T]): The dataset to wrap.
        cache_path (str | Path): Directory where cached samples will be stored.

    Methods:
        __len__(): Returns the number of samples in the dataset.
        __getitem__(index: int) -> Any: Retrieves a sample from cache or the original dataset.
    """

    def __init__(self, dataset: Dataset[T_co], cache_path: str | Path) -> None:
        self.dataset = dataset

        self.manager = Manager()
        self.cache = self.manager.dict()

        self.cache_path = Path(cache_path)
        self.cache_path.mkdir(exist_ok=True, parents=True)

    def __len__(self) -> int:
        return len(self.dataset)  # type: ignore[arg-type]

    def __getitem__(self, index: int) -> Any:
        if index not in self.cache:
            sample = self.dataset[index]

            self.cache[index] = self.cache_path / f"cache-{uuid.uuid4()}.pkl"
            with open(self.cache[index], "wb") as f:
                pkl.dump(sample, f)
        else:
            with open(self.cache[index], "rb") as f:
                sample = pkl.load(f)

        return sample



from pathlib import Path
import pickle as pkl

import optuna


def objective(x1: float, x2: float) -> float:
    return (x1**2) - (4.0 * x1) + (x2**2) - x2 - (x1 * x2)


if __name__ == "__main__":
    # ==========================================================
    # normal
    db_path = Path("./test_normal.db")
    study_name = "test_study_normal"
    url = f"sqlite:///{db_path}"

    if db_path.exists():
        db_path.unlink()

    storage = optuna.storages.RDBStorage(url=url)
    study = optuna.create_study(
        direction="minimize", sampler=optuna.samplers.TPESampler(seed=0), storage=storage, study_name=study_name
    )

    for _ in range(30):
        trial = study.ask()
        y = objective(x1=trial.suggest_float("x1", 0.0, 1.0), x2=trial.suggest_float("x2", 0.0, 1.0))
        frozentrial = study.tell(trial, y)
        study._log_completed_trial([y], frozentrial.number, frozentrial.params)

    with open("./test_notmal.pkl", "wb") as f:
        pkl.dump(study, f)

    # ==========================================================
    # resume
    db_path = Path("./test_resume.db")
    study_name = "test_study_resume"
    url = f"sqlite:///{db_path}"

    if db_path.exists():
        db_path.unlink()

    storage = optuna.storages.RDBStorage(url=url)
    study = optuna.create_study(
        direction="minimize", sampler=optuna.samplers.TPESampler(seed=0), storage=storage, study_name=study_name
    )

    for _ in range(15):
        trial = study.ask()
        y = objective(x1=trial.suggest_float("x1", 0.0, 1.0), x2=trial.suggest_float("x2", 0.0, 1.0))
        frozentrial = study.tell(trial, y)
        study._log_completed_trial([y], frozentrial.number, frozentrial.params)

    # resume
    print("resume after 14 trials")
    study = optuna.load_study(study_name=study_name, sampler=optuna.samplers.TPESampler(seed=0), storage=url)

    for _ in range(15):
        trial = study.ask()
        y = objective(x1=trial.suggest_float("x1", 0.0, 1.0), x2=trial.suggest_float("x2", 0.0, 1.0))
        frozentrial = study.tell(trial, y)
        study._log_completed_trial([y], frozentrial.number, frozentrial.params)

    with open("./test_resume.pkl", "wb") as f:
        pkl.dump(study, f)

    # ==========================================================
    # compare
    with open("./test_resume.pkl", "rb") as f:
        study_resume = pkl.load(f)
    with open("./test_notmal.pkl", "rb") as f:
        study_normal = pkl.load(f)

    try:
        assert study_resume.best_value == study_normal.best_value, (
            f"resume: {study_resume.best_value}, normal: {study_normal.best_value}"
        )
    except AssertionError as e:
        print(e)



import numpy.typing as npt

from unittest.mock import patch

import numpy as np

import pytest

from aiaccel.hpo.algorithms.nelder_mead_algorithm import (
    NelderMeadAlgorism,
    NelderMeadEmptyError,
    UnexpectedVerticesUpdateError,
)


@pytest.fixture
def search_space() -> dict[str, tuple[float, float]]:
    return {"x": (-5.0, 5.0), "y": (-5.0, 5.0)}


@pytest.fixture
def nm(search_space: dict[str, tuple[float, float]]) -> NelderMeadAlgorism:
    return NelderMeadAlgorism(dimensions=len(search_space), block=False)


@pytest.fixture
def vertices() -> list[npt.NDArray[np.float64]]:
    return list(np.array([[0.6, 0.7], [0.8, 0.9], [0.7, 0.9]]))


@pytest.fixture
def values() -> list[float]:
    return [5.0, 3.0, 7.0]


def test_waiting_for_result_empty(nm: NelderMeadAlgorism) -> None:
    nm.vertices, nm.values = [], []

    # queue is Empty
    result_value = next(nm._wait_for_result())
    assert result_value is None


def test_waiting_for_results_empty(nm: NelderMeadAlgorism) -> None:
    nm.vertices, nm.values = [], []

    # queue is Empty
    result_value = next(nm._wait_for_results(2))
    assert result_value is None


def test_waiting_for_result(nm: NelderMeadAlgorism) -> None:
    nm.vertices, nm.values = [], []

    # queue is not Empty
    expected_value = 1.0
    nm.put_value(np.zeros(2), expected_value)
    with pytest.raises(StopIteration) as e:
        next(nm._wait_for_result())
    assert e.value.value == expected_value


def test_waiting_for_results(nm: NelderMeadAlgorism) -> None:
    nm.vertices, nm.values = [], []

    # queue is not Empty
    expected_value1 = 1.0
    nm.put_value(np.zeros(2), expected_value1)
    expected_value2 = 2.0
    nm.put_value(np.zeros(2), expected_value2)
    with pytest.raises(StopIteration) as e:
        next(nm._wait_for_results(2))
    result_values = e.value.value[1]
    assert result_values == [expected_value1, expected_value2]


def test_waiting_for_results_enqueue_update(nm: NelderMeadAlgorism) -> None:
    with patch(
        "aiaccel.hpo.optuna.samplers.nelder_mead_sampler.NelderMeadAlgorism._collect_enqueued_results"
    ) as mock_iter:
        mock_iter.side_effect = [([np.array([-1.0, 0.0])], [0.5])]

        nm.vertices, nm.values = [], []

        # queue is not Empty
        expected_value1 = 1.0
        nm.put_value(np.zeros(2), expected_value1)
        expected_value2 = 2.0
        nm.put_value(np.zeros(2), expected_value2)
        with pytest.raises(UnexpectedVerticesUpdateError):
            next(nm._wait_for_results(2))


def test_collect_enqueued_results_empty(nm: NelderMeadAlgorism) -> None:
    # queue is Empty
    result_vertices, result_values = nm._collect_enqueued_results()
    assert result_vertices == []
    assert result_values == []


def test_collect_enqueued_results(nm: NelderMeadAlgorism) -> None:
    # queue is not Empty
    expected_vertex1 = np.array([1.0, 2.0])
    expected_value1 = 1.0
    nm.put_value(expected_vertex1, expected_value1, enqueue=True)
    expected_vertex2 = np.array([3.0, 4.0])
    expected_value2 = 2.0
    nm.put_value(expected_vertex2, expected_value2, enqueue=True)

    result_vertices, result_values = nm._collect_enqueued_results()
    assert np.array_equal(result_vertices, [expected_vertex1, expected_vertex2])
    assert result_values == [expected_value1, expected_value2]


def test_initialize(search_space: dict[str, tuple[float, float]], nm: NelderMeadAlgorism) -> None:
    for _ in range(len(search_space) + 1):
        random_xs = nm.get_vertex()
        for random_x, distribution in zip(random_xs, search_space.values(), strict=False):
            assert isinstance(random_x, float) and distribution[0] <= random_x <= distribution[1]

    with pytest.raises(NelderMeadEmptyError):
        nm.get_vertex()


def test_initialize_enqueued1(search_space: dict[str, tuple[float, float]], nm: NelderMeadAlgorism) -> None:
    # enqueued
    enqueued_vertex = np.array([0.6, 0.7])
    enqueued_value = 1.0
    nm.put_value(enqueued_vertex, enqueued_value, enqueue=True)

    for _ in range(len(search_space)):
        random_xs = nm.get_vertex()
        for random_x, distribution in zip(random_xs, search_space.values(), strict=False):
            assert isinstance(random_x, float) and distribution[0] <= random_x <= distribution[1]

    assert np.array_equal(nm.vertices, [enqueued_vertex])
    assert nm.values == [enqueued_value]

    with pytest.raises(NelderMeadEmptyError):
        nm.get_vertex()


def test_initialize_enqueued2(nm: NelderMeadAlgorism) -> None:
    # enqueued
    enqueud_vertices = [np.array([0.6, 0.7]), np.array([0.7, 0.8]), np.array([0.9, 1.0]), np.array([0.8, 0.9])]
    enqueud_values = [1.0, 4.0, 2.0, 3.0]
    expected_vertex = np.array([0.7, 0.8])

    for enqueued_vertex, enqueued_value in zip(enqueud_vertices, enqueud_values, strict=False):
        nm.put_value(enqueued_vertex, enqueued_value, enqueue=True)

    x = nm.get_vertex()
    assert np.allclose(x, expected_vertex)


@pytest.mark.parametrize(
    "expected_results",
    [
        pytest.param([{"vertex": [0.7, 0.7], "value": None, "enqueue": False}], id="reflect"),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 4.0, "enqueue": False},
                {"vertex": [0.9, 0.9], "value": None, "enqueue": False},
            ],
            id="reflect -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 2.0, "enqueue": False},
                {"vertex": [0.7, 0.6], "value": 1.0, "enqueue": False},
                {"vertex": [0.9, 0.8], "value": None, "enqueue": False},
            ],
            id="reflect -> expand -> fe < fr -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 2.0, "enqueue": False},
                {"vertex": [0.7, 0.6], "value": 3.0, "enqueue": False},
                {"vertex": [0.9, 0.9], "value": None, "enqueue": False},
            ],
            id="reflect -> expand -> else (fe > fr) -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 6.0, "enqueue": False},
                {"vertex": [0.7, 0.75], "value": 5.5, "enqueue": False},
                {"vertex": [0.7, 0.85], "value": None, "enqueue": False},
            ],
            id="reflect -> outside_contract -> foc <= fr -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 6.0, "enqueue": False},
                {"vertex": [0.7, 0.75], "value": 7.0, "enqueue": False},
                {"vertex": [0.7, 0.8], "value": 1.0, "enqueue": False},
                {"vertex": [0.75, 0.9], "value": 2.0, "enqueue": False},
                {"vertex": [0.65, 0.8], "value": None, "enqueue": False},
            ],
            id="reflect -> outside_contract -> shrink -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 8.0, "enqueue": False},
                {"vertex": [0.7, 0.85], "value": 6.0, "enqueue": False},
                {"vertex": [0.7, 0.75], "value": None, "enqueue": False},
            ],
            id="reflect -> inside_contract -> fic < self.values[-1] -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 8.0, "enqueue": False},
                {"vertex": [0.7, 0.85], "value": 8.5, "enqueue": False},
                {"vertex": [0.7, 0.8], "value": 1.0, "enqueue": False},
                {"vertex": [0.75, 0.9], "value": 2.0, "enqueue": False},
                {"vertex": [0.65, 0.8], "value": None, "enqueue": False},
            ],
            id="reflect -> inside_contract -> else (fic > self.values[-1]) -> shrink -> reflect",
        ),
        pytest.param(
            [
                {"vertex": [0.7, 0.7], "value": 4.0, "enqueue": False},
                {"vertex": [0.6, 0.3], "value": 2.0, "enqueue": True},
                {"vertex": [0.7, 0.5], "value": None, "enqueue": False},
            ],
            id="reflect -> UnexpectedVerticesUpdateError -> reflect",
        ),
    ],
)
def test_compare_results(
    vertices: list[npt.NDArray[np.float64]],
    values: list[float],
    nm: NelderMeadAlgorism,
    expected_results: list[dict[str, list[float] | float | bool]],
) -> None:
    for vertex, value in zip(vertices, values, strict=False):
        nm.put_value(vertex, value, True)

    # main loop
    for expected_result in expected_results:
        expected_vertex, expected_value, enqueued = expected_result.values()

        if not enqueued:
            x = nm.get_vertex()
            assert np.allclose(x, expected_vertex)

        if not isinstance(expected_value, float):
            break

        assert isinstance(enqueued, bool)
        nm.put_value(np.array(expected_vertex), expected_value, enqueued)



from typing import Any

import argparse
from concurrent.futures import FIRST_COMPLETED, ThreadPoolExecutor, wait
from importlib import resources
import json
from pathlib import Path
import shlex
import subprocess

from hydra.utils import instantiate
from omegaconf import OmegaConf as oc  # noqa: N813

from optuna.trial import Trial

from aiaccel.config import load_config, print_config, resolve_inherit


def main() -> None:
    """
    Main function to execute the hyperparameter optimization process using a Dask cluster.
    This function parses command-line arguments, loads the configuration,
    sets up the Dask client, and runs the optimization trials in a distributed manner.

    Command-line arguments:
        - --config (str, optional): Path to the configuration file.
        - --resume (bool, optional): Flag to resume from the previous study.
        - --resumable (bool, optional): Flag to make the study resumable by setting appropriate storage.

    Usage:
        - Start a new study:
            python -m aiaccel.hpo.apps.optimize --config config.yaml
        - Resume from the previous study:
            python -m aiaccel.hpo.apps.optimize --config config.yaml --resume
        - Make the study resumable:
            python -m aiaccel.hpo.apps.optimize --config config.yaml --resumable

    Config file (yaml) example:
        ~~~ yaml
        study:
          _target_: optuna.create_study
          direction: minimize
          storage:
            _target_: optuna.storages.InMemoryStorage
          study_name: aiaccel_study
          load_if_exists: false

        cluster:
          _target_: distributed.Client
          n_workers: 4
          threads_per_worker: 1

        sampler:
          _target_: optuna.samplers.TPESampler
          seed: 0

        params:
          _convert_: partial
          _target_: aiaccel.hpo.optuna.hparams_manager.HparamsManager
          x1: [0, 1]
          x2:
            _target_: aiaccel.apps.optimize.Float
            low: 0.0
            high: 1.0
            log: false

        n_trials: 30
        n_max_jobs: 4
        ~~~
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("command", nargs="+")

    parser.add_argument("--config", help="Configuration file path", default=None)
    parser.add_argument("--resume", action="store_true", default=False)
    parser.add_argument("--resumable", action="store_true", default=False)

    args, unk_args = parser.parse_known_args()

    with resources.as_file(resources.files("aiaccel.hpo.apps.config") / "default.yaml") as path:
        default_config = oc.load(path)
    config = oc.merge(default_config, load_config(args.config) if args.config is not None else {})
    config = oc.merge(config, oc.from_cli(unk_args))

    if (args.resumable or args.resume) and ("storage" not in config.study or args.config is None):
        with resources.as_file(resources.files("aiaccel.hpo.apps.config") / "resumable.yaml") as path:
            config = oc.merge(config, path)

    if args.resume:
        config.study.load_if_exists = True

    print_config(config)

    config = resolve_inherit(config)

    work_dir = Path.cwd()
    work_dir.mkdir(parents=True, exist_ok=True)

    study = instantiate(config.study)
    params = instantiate(config.params)

    futures: dict[Any, tuple[Trial, str]] = {}
    submitted_job_count = 0
    finished_job_count = 0

    with ThreadPoolExecutor(config.n_max_jobs) as pool:
        while finished_job_count < config.n_trials:
            active_jobs = len(futures.keys())
            available_slots = max(0, config.n_max_jobs - active_jobs)

            # Submit job in ThreadPoolExecutor
            for _ in range(min(available_slots, config.n_trials - submitted_job_count)):
                trial = study.ask()
                out_filename = config.out_filename_template.format(**vars())

                future = pool.submit(
                    subprocess.run,
                    shlex.join(args.command).format(
                        job_name=config.job_name_template.format(**vars()),
                        out_filename=out_filename,
                        **params.suggest_hparams(trial),
                    ),
                    shell=True,
                )

                futures[future] = trial, out_filename
                submitted_job_count += 1

            # Get result from out_filename and tell
            done_features, _ = wait(futures.keys(), return_when=FIRST_COMPLETED)
            for future in done_features:
                trial, out_filename = futures.pop(future)

                with open(out_filename) as f:
                    y = json.load(f)

                frozentrial = study.tell(trial, y)
                study._log_completed_trial(y if isinstance(y, list) else [y], frozentrial.number, frozentrial.params)
                finished_job_count += 1


if __name__ == "__main__":
    main()



from pathlib import Path

from omegaconf import OmegaConf as oc  # noqa: N813

import pytest

from aiaccel.config.config import load_config, pathlib2str_config, print_config, resolve_inherit


def test_load_config() -> None:
    loaded_config = load_config(Path(__file__).parent / "test_conf.yaml")
    expected_config = {
        "A": [{"_inherit_": ["${B}", "${C}"], "AA": "aa"}, {"AAA": "aaa"}],
        "B": {"AA": "dummy", "BB": "bb"},
        "C": {"CC": "cc"},
        "D": {"_inherit_": "E"},
        "E": {"EE": "ee"},
    }

    assert loaded_config == expected_config


def test_resolve_inherit() -> None:
    loaded_config = oc.create(
        {
            "A": [{"_inherit_": ["${B}", "${C}"], "AA": "aa"}, {"AAA": "aaa"}],
            "B": {"AA": "dummy", "BB": "bb"},
            "C": {"CC": "cc"},
            "D": {"_inherit_": "${E}"},
            "E": {"EE": "ee"},
        }
    )
    resolved_config = resolve_inherit(loaded_config)
    expected_config = {
        "A": [{"CC": "cc", "AA": "aa", "BB": "bb"}, {"AAA": "aaa"}],
        "B": {"AA": "dummy", "BB": "bb"},
        "C": {"CC": "cc"},
        "D": {"EE": "ee"},
        "E": {"EE": "ee"},
    }

    assert resolved_config == expected_config


def test_print_config(capfd: pytest.CaptureFixture[str]) -> None:
    conf = oc.create({"foo": {"bar": [1, 2, 3]}})
    print_config(conf)

    stdout, _ = capfd.readouterr()

    # with open(Path(__file__).parent / "test_config_assets" / "print_config.txt", "w") as f:
    #     f.write(stdout)  # noqa: ERA001

    with open(Path(__file__).parent / "test_config_assets" / "print_config.txt") as f:
        stdout_target = f.read()

    assert stdout == stdout_target


def test_pathlib2str_config() -> None:
    src_conf = oc.create({"foo": {"bar": Path("test/path")}})
    dst_conf = pathlib2str_config(src_conf)

    assert isinstance(dst_conf.foo.bar, str)



from argparse import ArgumentParser
import logging
import os
from pathlib import Path

from hydra.utils import instantiate
from omegaconf import OmegaConf as oc  # noqa: N813

import lightning as lt

from aiaccel.config import (
    load_config,
    overwrite_omegaconf_dumper,
    pathlib2str_config,
    print_config,
    resolve_inherit,
)
from aiaccel.config.git import collect_git_status_from_config, print_git_status

logger = logging.getLogger(__name__)


def main() -> None:
    overwrite_omegaconf_dumper()

    parser = ArgumentParser()
    parser.add_argument("config", type=str, help="Config file in YAML format")
    args, unk_args = parser.parse_known_args()

    # load config
    config = oc.merge(
        load_config(
            args.config,
            {
                "config_path": args.config,
                "working_directory": str(Path(args.config).parent.resolve()),
                "base_config_path": str(Path(__file__).parent / "config"),
            },
        ),
        oc.from_cli(unk_args),
    )

    if int(os.environ.get("OMPI_COMM_WORLD_RANK", 0)) == 0 and int(os.environ.get("RANK", 0)) == 0:
        print_config(config)

        status_list = collect_git_status_from_config(config)
        print_git_status(status_list)

    config = resolve_inherit(config)

    # build trainer
    trainer: lt.Trainer = instantiate(config.trainer)

    # save config
    if trainer.is_global_zero:
        Path(config.working_directory).mkdir(parents=True, exist_ok=True)
        merged_config_path = Path(config.working_directory) / "merged_config.yaml"

        with open(merged_config_path, "w") as f:
            oc.save(pathlib2str_config(config), f)

    # start training
    trainer.fit(
        model=instantiate(config.task),
        datamodule=instantiate(config.datamodule),
    )


if __name__ == "__main__":
    main()



import argparse


def main(x1: float, x2: float) -> None:
    y = (x1**2) - (4.0 * x1) + (x2**2) - x2 - (x1 * x2)
    print(y)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--x1", type=float)
    parser.add_argument("--x2", type=float)

    args, unk_args = parser.parse_known_args()

    main(args.x1, args.x2)



from aiaccel.torch.lightning.abci_environment import ABCIEnvironment
from aiaccel.torch.lightning.opt_lightning_module import OptimizerConfig, OptimizerLightningModule, build_param_groups

__all__ = ["ABCIEnvironment", "OptimizerConfig", "OptimizerLightningModule", "build_param_groups"]



#! /usr/bin/env python3

from typing import Any

from argparse import ArgumentParser, Namespace
from functools import partial
from multiprocessing import Pool
from pathlib import Path
import shlex
import subprocess
import sys

import yaml

default_config_yaml = """\
script_prologue: |
    echo Hostname: $(hostname)

    export NVIDIA_VISIBLE_DEVICES=all

cpu:
    job: "{command}"

cpu-array:
    n_tasks_per_proc: 128
    n_procs: 24
    job: "{command}"

gpu:
    job: "{command}"

gpu-array:
    n_tasks_per_proc: 128
    n_procs: 8
    job: "CUDA_VISIBLE_DEVICES=$(( LOCAL_PROC_INDEX % {args.n_procs_per_job} )) {command}"

mpi:
    job: |
        mpirun -np {args.n_procs} \\
            {command}

train:
    job: |
        mpirun -np {args.n_gpus} \\
            -x MAIN_ADDR=$(hostname -i) \\
            -x MAIN_PORT=3000 \\
            -x COLUMNS=120 \\
            -x PYTHONUNBUFFERED=true \\
            {command}
"""  # noqa: E501


def dispatch_job(mode: str, args: Namespace, config: Any) -> None:
    # Prepare the job script and arguments
    job = config[mode]["job"].format(command=shlex.join(args.command), args=args)

    if mode in ["cpu-array", "gpu-array"]:
        job = f"TASK_STEPSIZE={args.n_tasks_per_proc} {job}"
        log_filename = f"{args.log_filename.with_suffix('')}.${{TASK_INDEX}}.log"
    else:
        log_filename = args.log_filename

    job_script = f"""\
#! /bin/bash

set -eE -o pipefail
trap 'exit $?' ERR EXIT  # at error and exit
trap 'echo 143' TERM  # at termination (by job scheduler)


{config["script_prologue"]}

{job} 2>&1 | tee {log_filename}
"""

    # Create the job script file, remove old status files, and run the job
    args.log_filename.parent.mkdir(exist_ok=True, parents=True)

    job_filename: Path = args.log_filename.with_suffix(".sh")
    with open(job_filename, "w") as f:
        f.write(job_script)

    if mode in ["cpu-array", "gpu-array"]:
        worker = partial(subprocess.run, shell=True, check=True)
        with Pool(processes=args.n_procs) as pool:
            cmd_list = []
            for ii in range(0, args.n_tasks, args.n_tasks_per_proc):
                cmd_list.append(f"TASK_INDEX={ii + 1} bash {job_filename}")

            for _ in pool.imap_unordered(worker, cmd_list):
                pass
    else:
        subprocess.run(f"bash {job_filename}", shell=True, check=True)


def main() -> None:
    # Load configuration (from the default YAML string)
    parser = ArgumentParser(add_help=False)
    parser.add_argument("--print_config", action="store_true")
    parser.add_argument("--config", type=Path)
    args, _ = parser.parse_known_args()

    if args.config is not None:
        with open(args.config) as f:
            config_yaml = f.read()
    else:
        config_yaml = default_config_yaml

    if args.print_config:
        print(config_yaml)
        sys.exit(0)

    config = yaml.safe_load(config_yaml)

    # Parse command-line arguments
    parser = ArgumentParser()
    parser.add_argument("--print_config", action="store_true")
    parser.add_argument("--config", type=Path)
    sub_parsers = parser.add_subparsers(dest="mode", required=True)

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument("--walltime", type=str)  # defined for compatibility
    parent_parser.add_argument("log_filename", type=Path)
    parent_parser.add_argument("command", nargs="+")

    sub_parser = sub_parsers.add_parser("cpu", parents=[parent_parser])
    sub_parser.add_argument("--n_tasks", type=int)
    sub_parser.add_argument("--n_tasks_per_proc", type=int, default=config["cpu-array"]["n_tasks_per_proc"])
    sub_parser.add_argument("--n_procs", type=int, default=config["cpu-array"]["n_procs"])

    sub_parser = sub_parsers.add_parser("gpu", parents=[parent_parser])
    sub_parser.add_argument("--n_tasks", type=int)
    sub_parser.add_argument("--n_tasks_per_proc", type=int, default=config["gpu-array"]["n_tasks_per_proc"])
    sub_parser.add_argument("--n_procs", type=int, default=config["gpu-array"]["n_procs"])

    sub_parser = sub_parsers.add_parser("mpi", parents=[parent_parser])
    sub_parser.add_argument("--n_procs", type=int, required=True)
    sub_parser.add_argument("--n_nodes", type=int)  # defined for compatibility

    sub_parser = sub_parsers.add_parser("train", parents=[parent_parser])
    sub_parser.add_argument("--n_gpus", type=int)

    args = parser.parse_args()
    mode = args.mode + "-array" if getattr(args, "n_tasks", None) is not None else args.mode

    for key in ["walltime", "n_nodes"]:
        if getattr(args, key, None) is not None:
            print(f"Argument '{key}' is defined for compatibility and will not be used in aiaccel-job local.")

    dispatch_job(mode, args, config)


if __name__ == "__main__":
    main()



from typing import Any

from collections.abc import KeysView
from pathlib import Path
from types import TracebackType

# This stub is just for passing mypy.
class Dataset:
    def __setitem__(self, arg: Any, value: Any) -> None: ...
    def __getitem__(self, key: Any) -> Any: ...

class Group:
    def create_dataset(self, name: str, shape: Any | None = None, dtype: Any | None = None) -> Dataset: ...
    def __getitem__(self, name: Any) -> Dataset: ...
    def items(self) -> list[tuple[str, Any]]: ...

class File:
    def __init__(
        self,
        name: str | Path,
        mode: str = "r",
        driver: str | None = None,
        comm: Any | None = None,
    ) -> None: ...
    def __enter__(self) -> File: ...
    def __exit__(
        self,
        ex_exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool: ...
    def keys(self) -> KeysView[str]: ...
    def __getitem__(self, key: str) -> Group | Dataset: ...
    def close(self) -> None: ...
    def create_group(self, name: str) -> Group: ...



from torch import nn

from torchvision import models


class SmallResNet50(nn.Sequential):
    def __init__(self, num_classes: int):
        super().__init__()

        self.base = models.resnet50(num_classes=num_classes)
        self.base.conv1 = nn.Conv2d(3, 64, 3, 1, 1, bias=False)
        self.base.maxpool = nn.Identity()
        self.base.fc = nn.Linear(2048, 10)



from collections.abc import Generator
import os
from pathlib import Path
import shutil
import sqlite3
import subprocess
import tempfile
import uuid

import pytest


@pytest.fixture
def temp_dir() -> Generator[Path]:
    """Create a temporary directory with test files and clean up afterwards"""
    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_path = Path(tmp_dir)
        original_dir = os.getcwd()
        os.chdir(tmp_dir)

        source_dir = Path(__file__).parent
        test_files = [
            "config.yaml",
            "objective_for_test.py",
            "execute_optuna_for_test.py",
            "config_multi.yaml",
            "objective_multi_for_test.py",
        ]

        for file_name in test_files:
            source_file = source_dir / file_name
            target_file = temp_path / file_name
            if source_file.exists():
                shutil.copy2(source_file, target_file)
                os.chmod(target_file, 0o755)
                print(f"\n=== Content of {file_name} ===")
                print((target_file).read_text())
                print("=" * 40)
            else:
                pytest.skip(f"Required test file {file_name} not found in {source_dir}")

        yield temp_path
        os.chdir(original_dir)


def get_trial_count(db_path: Path, study_name: str) -> int:
    """Get the number of trials from the SQLite database for a specific study

    Args:
        db_path (Path): Path to the SQLite database
        study_name (str): Name of the study to query

    Returns:
        int: Number of trials for the specified study
    """
    if not db_path.exists():
        raise AssertionError("Database file does not exist")
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT COUNT(*) FROM trials
        WHERE study_id = (SELECT study_id FROM studies WHERE study_name = ?)
    """,
        (study_name,),
    )
    count = cursor.fetchone()[0]
    conn.close()
    if count is None:
        return 0
    return int(count)


def get_trial_values(db_path: Path, study_name: str) -> list[float]:
    """Get the values of all completed trials for a specific study

    Args:
        db_path (Path): Path to the SQLite database
        study_name (str): Name of the study to query

    Returns:
        list[float]: List of trial values in order of trial number
    """
    if not db_path.exists():
        raise AssertionError("Database file does not exist")
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT trial_values.value
        FROM trials
        JOIN studies ON trials.study_id = studies.study_id
        JOIN trial_values ON trials.trial_id = trial_values.trial_id
        WHERE studies.study_name = ?
            AND trials.state = 'COMPLETE'
        ORDER BY trials.number
    """,
        (study_name,),
    )
    # Fetch all values from the query result

    values = [row[0] for row in cursor.fetchall()]
    conn.close()
    return values


def modify_config(config_path: Path, study_name: str, n_trials: int, db_name: str) -> Path:
    """Modify config file with new study name and number of trials"""
    with open(config_path) as f:
        content = f.read()
    content = content.replace("study_name: my_study", f"study_name: {study_name}")
    content = content.replace("n_trials: 30", f"n_trials: {n_trials}")

    if db_name:
        content = content.replace("url: sqlite:///aiaccel_storage.db", f"url: sqlite:///{db_name}")
    new_config_path = config_path.parent / f"config_{study_name}.yaml"
    with open(new_config_path, "w") as f:
        f.write(content)
    print(f"\nnew_config_path: {new_config_path}")
    return new_config_path


def _run_aiaccel(cfg: Path, objective: str, extra_args: list[str]) -> None:
    command = ["aiaccel-hpo", "optimize", "--config", str(cfg)]
    command += extra_args
    command += ["--", "bash", "-c", f"python {objective} --x1={{x1}} --x2={{x2}} > {{out_filename}}"]

    subprocess.run(
        command,
        check=True,
    )


def test_optimization_consistency(temp_dir: Path) -> None:
    """Test that split execution (resumable + resume) gives same results as normal execution."""
    # Use different database files for normal and split execution
    normal_db = "normal_storage.db"
    split_db = "split_storage.db"

    # Normal execution with 30 trials
    study_name_normal = f"test_study_{uuid.uuid4().hex[:8]}"
    normal_config = modify_config(temp_dir / "config.yaml", study_name_normal, 30, normal_db)

    _run_aiaccel(normal_config, "objective_for_test.py", [])

    normal_results = get_trial_values(temp_dir / normal_db, study_name_normal)
    assert len(normal_results) == 30, "Normal execution should have 30 trials"
    normal_best = min(normal_results)

    trial_count = get_trial_count(temp_dir / normal_db, study_name_normal)
    assert trial_count == 30

    # Split execution with 15 + 15 trials
    study_name_split = f"test_study_{uuid.uuid4().hex[:8]}"
    split_config = modify_config(temp_dir / "config.yaml", study_name_split, 15, split_db)

    # First 15 trials
    _run_aiaccel(split_config, "objective_for_test.py", ["--resumable"])

    trial_count = get_trial_count(temp_dir / split_db, study_name_split)

    assert trial_count == 15

    # Second 15 trials
    _run_aiaccel(split_config, "objective_for_test.py", ["--resume"])

    trial_count = get_trial_count(temp_dir / split_db, study_name_split)
    assert trial_count == 30

    # optuna resume
    subprocess.run(["python", "execute_optuna_for_test.py"], check=True)
    # get best value
    normal_result = get_trial_values(temp_dir / "test_normal.db", "test_study_normal")
    normal_expected_best = min(normal_result)
    resume_result = get_trial_values(temp_dir / "test_resume.db", "test_study_resume")
    resume_expected_best = min(resume_result)

    # Compare results
    resume_results = get_trial_values(temp_dir / split_db, study_name_split)
    resume_best = min(resume_results)
    assert len(resume_results) == 30, "Split execution should have 30 trials"
    ## optuna's best value VS aiaccel's best value (normal execution)
    assert abs(normal_expected_best - normal_best) < 1e-6, (
        f"Best values differ: normal={normal_best}, optuna={normal_expected_best}"
    )
    ## optuna's best value VS aiaccel's best value (resume execution)
    assert abs(resume_expected_best - resume_best) < 1e-6, (
        f"Best values differ: resume={resume_best}, optuna={resume_expected_best}"
    )


def test_normal_execution(temp_dir: Path) -> None:
    """Test normal execution without resume functionality"""

    db_name = "normal_test.db"
    study_name = f"test_study_{uuid.uuid4().hex[:8]}"
    config_path = modify_config(temp_dir / "config.yaml", study_name, 30, db_name)

    _run_aiaccel(config_path, "objective_for_test.py", [])

    trial_count = get_trial_count(temp_dir / db_name, study_name)
    assert trial_count == 30


def test_resumable_execution(temp_dir: Path) -> None:
    """Test execution with `--resumable`
    This should run only the first half of the trials and save the state to a database file.
    """

    db_name = "resumable_test.db"
    study_name = f"test_study_{uuid.uuid4().hex[:8]}"
    config_path = modify_config(temp_dir / "config.yaml", study_name, 15, db_name)

    _run_aiaccel(config_path, "objective_for_test.py", ["--resumable"])

    db_path = temp_dir / db_name
    assert db_path.exists(), "Database file was not created"
    trial_count = get_trial_count(db_path, study_name)
    assert trial_count == 15


def test_resume_execution(temp_dir: Path) -> None:
    """Test the resume functionality of the optimization process.

    This test verifies that the optimization process can be resumed correctly
    from a saved state. It performs the following steps:
    1. Runs the optimization process with the `--resumable` flag and checks
       that the correct number of trials are completed.
    2. Resumes the optimization process with the `--resume` flag and checks
       that the remaining trials are completed correctly.

    Args:
        temp_dir (pathlib.Path): Temporary directory for test files.

    Raises:
        AssertionError: If the number of trials after each run does not match
                        the expected values.
    """

    db_name = "resume_test.db"
    study_name = f"test_study_{uuid.uuid4().hex[:8]}"
    config_path = modify_config(temp_dir / "config.yaml", study_name, 15, db_name)

    _run_aiaccel(config_path, "objective_for_test.py", ["--resumable"])

    db_path = temp_dir / db_name
    trial_count = get_trial_count(db_path, study_name)
    assert trial_count == 15

    _run_aiaccel(config_path, "objective_for_test.py", ["--resume"])

    trial_count = get_trial_count(db_path, study_name)
    assert trial_count == 30


def test_multi_execution(temp_dir: Path) -> None:
    db_name = "multi_test.db"
    study_name = f"test_study_{uuid.uuid4().hex[:8]}"
    config_path = modify_config(temp_dir / "config_multi.yaml", study_name, 15, db_name)

    _run_aiaccel(config_path, "objective_multi_for_test.py", ["--resumable"])

    db_path = temp_dir / db_name
    trial_count = get_trial_count(db_path, study_name)
    assert trial_count == 15

    _run_aiaccel(config_path, "objective_multi_for_test.py", ["--resume"])

    trial_count = get_trial_count(db_path, study_name)
    assert trial_count == 30



import numpy as np
from scipy.optimize import linear_sum_assignment as scipy_linear_sum_assignment

import torch


def linear_sum_assignment(cost_matrix: torch.Tensor, maximize: bool = False) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Solve the linear sum assignment problem for a batch of cost matrices.

    Args:
        cost_matrix (torch.Tensor): A tensor of shape (..., m, n)
            representing the cost matrix for each assignment problem.
        maximize (bool): If True, the problem is treated as a maximization problem.
            If False, it is treated as a minimization problem. Defaults to False.
    Returns:
        tuple: A tuple containing two tensors:
            - row_indices: Indices of the rows assigned to each column.
            - col_indices: Indices of the columns assigned to each row.
    """

    assert cost_matrix.ndim >= 2, "cost_matrix must have at least 2 dimensions"

    *batch_shape, m, n = cost_matrix.shape

    row_ind_list, col_ind_list = [], []
    for cm in cost_matrix.reshape(-1, m, n).cpu().numpy():
        row_ind, col_ind = scipy_linear_sum_assignment(cm, maximize=maximize)

        row_ind_list.append(row_ind)
        col_ind_list.append(col_ind)

    row_indices = torch.from_numpy(np.stack(row_ind_list).reshape(*batch_shape, -1)).to(cost_matrix.device)
    col_indices = torch.from_numpy(np.stack(col_ind_list).reshape(*batch_shape, -1)).to(cost_matrix.device)

    return row_indices, col_indices



from dataclasses import dataclass
import importlib.util
import os
from pathlib import Path
import subprocess

from omegaconf import DictConfig, ListConfig

__all__ = [
    "PackageGitStatus",
    "collect_git_status_from_config",
    "print_git_status",
]


@dataclass
class PackageGitStatus:
    """
    Represents the Git status of a package.

    Attributes:
        package_name (str): The name of the package.
        commit_id (str): The current Git commit ID of the repository.
        status (list[str]): A list of uncommitted files in the repository.
    """

    package_name: str
    commit_id: str
    status: list[str]

    def ready(self) -> bool:
        """
        Determines if there are no uncommitted changes.

        Returns:
            bool: True if there are no uncommitted files, otherwise False.
        """

        return len(self.status) == 0


def collect_git_status_from_config(config: DictConfig | ListConfig) -> list[PackageGitStatus]:
    """
    Collects the Git status of packages specified in the given configuration.

    Args:
        config (DictConfig | ListConfig): The configuration containing package references.

    Returns:
        list[PackageGitStatus]: A list of `PackageGitStatus` objects representing
                                the Git status of the detected packages.
    """

    status_list = []

    package_names = collect_target_packages(config)
    package_names.sort()

    for package_name in package_names:
        status = get_git_status(package_name)

        if status is not None:
            status_list.append(status)

    return status_list


def print_git_status(status: PackageGitStatus | list[PackageGitStatus]) -> None:
    """
    Prints the Git status of a package or a list of packages.

    Args:
        status (PackageGitStatus | list[PackageGitStatus]): The Git status to print.
    """

    status_list = status if isinstance(status, list) else [status]

    for status in status_list:
        print(f"{status.package_name} @ {status.commit_id}")
        for st in status.status:
            print(f"  {st}")


def get_git_status(package_name: str) -> PackageGitStatus | None:
    """
    Retrieves the Git status of a given package.

    Args:
        package_name (str): The name of the package to check.

    Returns:
        PackageGitStatus | None: A `PackageGitStatus` object if the package is found
                                 and under Git control, otherwise None.
    """

    # get package location
    spec = importlib.util.find_spec(package_name)

    if spec is None:
        return None

    if spec.origin is not None:
        module_path = Path(spec.origin).parent
    elif spec.submodule_search_locations is not None:
        module_path = Path(os.path.abspath(spec.submodule_search_locations[0]))
    else:
        return None

    # get repository path
    result = subprocess.run(["git", "rev-parse", "--show-toplevel"], cwd=module_path, capture_output=True, text=True)
    if result.returncode != 0:
        return None

    repository_path = result.stdout.splitlines()[0]

    # get commit id
    result = subprocess.run(["git", "rev-parse", "HEAD"], cwd=repository_path, capture_output=True, text=True)
    commit_id = result.stdout.splitlines()[0]

    # check git status
    result = subprocess.run(["git", "status", "-s"], cwd=repository_path, capture_output=True, text=True)
    status = result.stdout.splitlines()

    return PackageGitStatus(package_name, commit_id, status)


def collect_target_packages(config: ListConfig | DictConfig) -> list[str]:
    """
    Extracts the names of target packages from the given configuration.

    Args:
        config (ListConfig | DictConfig): The configuration to process.

    Returns:
        list[str]: A list of package names extracted from the configuration.
    """

    target_packages = set()

    def inner_func(_config: ListConfig | DictConfig) -> None:
        if isinstance(_config, DictConfig):
            for key, value in _config.items():
                if key == "_target_":
                    package_name, *_ = value.split(".")
                    target_packages.add(package_name)

                inner_func(value)

        elif isinstance(_config, ListConfig):
            for item in _config:
                inner_func(item)

    inner_func(config)

    return list(target_packages)



from aiaccel.config.config import (
    load_config,
    overwrite_omegaconf_dumper,
    pathlib2str_config,
    print_config,
    resolve_inherit,
)
from aiaccel.config.git import PackageGitStatus, collect_git_status_from_config, print_git_status

__all__ = [
    "load_config",
    "overwrite_omegaconf_dumper",
    "pathlib2str_config",
    "print_config",
    "resolve_inherit",
    "PackageGitStatus",
    "collect_git_status_from_config",
    "print_git_status",
]



import optuna

from aiaccel.hpo.optuna.hparams import (
    Categorical,
    Const,
    Float,
    Int,
)


def test_const() -> None:
    const = Const(value=0.5)
    assert const(trial=None, name="x1") == 0.5


def test_float() -> None:
    suggest_float = Float(low=0.0, high=1.0, step=None, log=False)
    trial = optuna.create_study().ask()

    assert isinstance(suggest_float(trial=trial, name="x2"), float)


def test_int() -> None:
    suggest_int = Int(low=0, high=10, step=1, log=False)
    trial = optuna.create_study().ask()

    assert isinstance(suggest_int(trial=trial, name="x3"), int)


def test_categorical() -> None:
    suggest_categorical = Categorical(choices=[0, 1, 2])
    trial = optuna.create_study().ask()

    assert suggest_categorical(trial=trial, name="x4") in [0, 1, 2]


def test_discrete_uniform() -> None:
    suggest_discrete_uniform = Float(low=0.0, high=1.0, step=0.1)
    trial = optuna.create_study().ask()

    assert isinstance(suggest_discrete_uniform(trial=trial, name="x5"), float)


def test_log_uniform() -> None:
    suggest_log_uniform = Float(low=0.1, high=1.0, log=True)
    trial = optuna.create_study().ask()

    assert isinstance(suggest_log_uniform(trial=trial, name="x6"), float)



import argparse


def main(x1: float, x2: float) -> list[float]:
    obj1 = (x1 - 2) ** 2 + (x2 - 1) ** 2
    obj2 = x1 + x2

    return [obj1, obj2]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--x1", type=float)
    parser.add_argument("--x2", type=float)

    args, unk_args = parser.parse_known_args()

    print(main(args.x1, args.x2))
